import { Bot, Schema, segment, Session } from 'koishi';
import { WeComBot } from './bot';
import {
  WecomChatBody,
  WecomEventResponse,
  WecomPicBody,
  WecomResponse,
  WeComUser,
} from './def';

export interface AdapterConfig {
  path?: string;
  token?: string;
  encodingAESKey?: string;
}

export const AdapterConfig: Schema<AdapterConfig> = Schema.object({
  path: Schema.string().description('企业微信回调路径。').default('/wecom'),
  token: Schema.string()
    .role('secret')
    .required()
    .description('应用消息上报 token。'),
  encodingAESKey: Schema.string()
    .role('secret')
    .required()
    .description('应用消息上播 AES 密钥。'),
});

export function adaptUser(user: WeComUser): Bot.User {
  return {
    userId: user.userid,
    username: user.name,
    nickname: user.name,
    avatar: user.avatar,
  };
}

export function adaptSession(bot: WeComBot, input: WecomEventResponse) {
  const { body } = input;
  if (!body) {
    return;
  }
  const session: Partial<Session> = {
    selfId: bot.selfId,
    messageId: body.MsgId?.toString(),
    wecom: input,
    userId: body.FromUserName,
    channelId: body.FromUserName,
    timestamp: body.CreateTime,
  };
  if (body.MsgType === 'event') {
    session.type = 'wecom/event';
    session.subtype = body.Event;
  } else {
    session.type = 'message';
    session.subtype = 'private';
    session.author = {
      userId: session.userId,
    };
    switch (body.MsgType) {
      case 'text':
        const textBody = body as WecomChatBody;
        session.content = textBody.Content;
        break;
      case 'image':
        const imageBody = body as WecomPicBody;
        session.content = segment('image', {
          url: imageBody.PicUrl,
        });
        break;
      default:
        return;
    }
  }
  return session;
}

export function dispatchSession(bot: WeComBot, message: WecomResponse<any>) {
  const payload = adaptSession(bot, message);
  if (!payload) return;
  const session = new Session(bot, payload);
  session.wecom = message;
  bot.adapter.dispatch(session);
}
